// Copyright (c) 2010 Andrew Lewis. All rights reserved.
// Use of this source code is governed the license in the LICENSE file.

#ifndef _INCLUDE_DATA_H_
#define _INCLUDE_DATA_H_

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#define MAX(x,y) ((INT32)(x) > (INT32)(y) ? (x) : (y))
#define MIN(x,y) ((INT32)(x) < (INT32)(y) ? (x) : (y))

#define INT8 char
#define UINT8 unsigned char
#define INT16 int
#define UINT16 unsigned int
#define INT32 int
#define UINT32 unsigned int
#define INT64 long int
#define UINT64 unsigned long int
#define INT128 __int128_t
#define UINT128 unsigned long long int

#ifndef HAVE_BOOLEAN
#define HAVE_BOOLEAN
typedef INT32 boolean;
#endif

#define TRUE 1
#define FALSE 0

#define UNION_BUFFER_UNDEFINED (-32768)
#define LIST_SENTINEL32 (0xFF000000)

typedef struct packed_image_data
{
  UINT8* samples;
  UINT16 width;
  UINT16 height;
} packed_image_data;
packed_image_data create_packed_image_data(UINT16 width, UINT16 height);
void destroy_packed_image_data(packed_image_data data);

typedef struct inverse_ycc_rgb_table
{
  UINT32** conversion_table;
  INT16* cr_r_tab;
  INT16* cb_b_tab;
  INT32* cr_g_tab;
  INT32* cb_g_tab;
} inverse_ycc_rgb_table;
inverse_ycc_rgb_table create_inverse_ycc_rgb_table();
void destroy_inverse_ycc_rgb_table(inverse_ycc_rgb_table table);

typedef UINT16* quantization_table;
quantization_table create_quantization_table();
void destroy_quantization_table(quantization_table table);

typedef struct interval_plane
{
  INT16* min;
  INT16* max;
  UINT16 width;
  UINT16 height;
} interval_plane;
interval_plane create_interval_plane(UINT16 width, UINT16 height);
void destroy_interval_plane(interval_plane plane);

typedef struct set_plane
{
  UINT32** sets;
  UINT16 width;
  UINT16 height;
} set_plane;
set_plane create_set_plane(packed_image_data original);
void destroy_set_plane(set_plane data);

typedef struct interval_component_image_data
{
  interval_plane y;
  interval_plane cb;
  interval_plane cr;
  UINT16 width;
  UINT16 height;
} interval_component_image_data;
interval_component_image_data create_interval_component_image_data(packed_image_data original, boolean chroma_subsampling);
void destroy_interval_component_image_data(interval_component_image_data data);
void set_maximum_chroma_intervals(interval_component_image_data data);
void copy_luma_data_into_component_data(set_plane luma_source, interval_component_image_data luma_destination, const UINT8 luma_shift);

typedef struct block_interval_component_image_data
{
  interval_plane y;
  interval_plane cb;
  interval_plane cr;
  UINT16 blocks_width;
  UINT16 blocks_height;
} block_interval_component_image_data;
block_interval_component_image_data create_block_interval_component_image_data(interval_component_image_data data);
void destroy_block_interval_component_image_data(block_interval_component_image_data data);

#define COMPUTATION_INPUT_JSAMPLE 1
#define COMPUTATION_IMMEDIATE_INTEGER 2
#define COMPUTATION_INTEGER_MULTIPLY 3
#define COMPUTATION_INTEGER_ADD 4
#define COMPUTATION_INTEGER_RSHIFT 5
#define COMPUTATION_ADD 6

typedef unsigned short int ComputationType;

typedef struct Computation
{
  ComputationType type;
  struct Computation* computation1;
  struct Computation* computation2;
  int constant;
  int component_index;

  int is_known;
  long int* value;

  long int* estimateMax;
  long int* estimateMin;
} Computation;
// TODO: Here we return/take a pointer to a computation rather than a computation on the stack. This is not consistent with the other data types.
Computation* create_computation(ComputationType type, Computation* c1, Computation* c2, int constant, int component_index);
void destroy_computation(Computation* computation);
Computation* clone_computation(Computation* computation);

typedef struct computation_plane
{
  Computation** computations;
  UINT16 width;
  UINT16 height;
  UINT16 input_width;
  UINT16 input_height;
} computation_plane;
computation_plane create_computation_plane(set_plane image, UINT16 input_width, UINT16 input_height);
void destroy_computation_plane(computation_plane image);

typedef struct block_status_summary
{
  UINT16 exact_block_count;
  UINT16 ambiguous_block_count;
  UINT16 infeasible_block_count;
  UINT16 impossible_block_count;
} block_status_summary;

#endif
