// Copyright (c) 2010 Andrew Lewis. All rights reserved.
// Use of this source code is governed the license in the LICENSE file.

#include "diagnosticinformation.h"
#include "forwardoperations.h"

void infeasible_block_information_output(interval_component_image_data downsampled_chroma_estimate, block_interval_component_image_data refined_block_data, const char* infeasible_block_file_name)
{
  UINT8 bx, by, bi, scanx, scany;
  INT32 minimum, maximum, average, average_interval_size;
  boolean infeasible = FALSE;

  FILE* information_file = fopen(infeasible_block_file_name, "w");
  if (information_file == NULL)
    return;

  for (by = 0; by < downsampled_chroma_estimate.y.height / 8; by++)
  {
    for (bx = 0; bx < downsampled_chroma_estimate.y.width / 8; bx++)
    {
      infeasible = (possibility_count(refined_block_data.y, bx, by) > possibilities_limit);
      minimum = (1 << 16);
      maximum = -(1 << 16);
      average = 0;
      average_interval_size = 0;
      for (bi = 0; bi < 64; bi++)
      {
        scanx = bx * 8 + (bi % 8);
        scany = by * 8 + (bi / 8);
        if (downsampled_chroma_estimate.y.min[scanx + scany * downsampled_chroma_estimate.y.width] < minimum)
          minimum = downsampled_chroma_estimate.y.min[scanx + scany * downsampled_chroma_estimate.y.width];
        if (downsampled_chroma_estimate.y.max[scanx + scany * downsampled_chroma_estimate.y.width] > maximum)
          maximum = downsampled_chroma_estimate.y.max[scanx + scany * downsampled_chroma_estimate.y.width];
        average += (downsampled_chroma_estimate.y.max[scanx + scany * downsampled_chroma_estimate.y.width] + downsampled_chroma_estimate.y.min[scanx + scany * downsampled_chroma_estimate.y.width]) / 2;
        average_interval_size += (downsampled_chroma_estimate.y.max[scanx + scany * downsampled_chroma_estimate.y.width] - downsampled_chroma_estimate.y.min[scanx + scany * downsampled_chroma_estimate.y.width]);
      }
      average /= 64;
      fprintf(information_file, "%s%d %d %d %d/64\n", infeasible ? "* " : "", minimum, average, maximum, average_interval_size);
    }
  }

  fclose(information_file);
}

INT32 quantize(INT32 value, INT32 qval)
{
  qval <<= 3;
  if (value < 0)
  {
    value = -value;
    value += (qval >> 1);
    value /= qval;
    value = -value;
  }
  else
  {
    value += (qval >> 1);
    value /= qval;
  }

  return value;
}

void substitute_fdct_infeasible_blocks(interval_component_image_data downsampled_chroma_estimate, block_interval_component_image_data refined_block_data, quantization_table luma_table, quantization_table chroma_table)
{
  UINT8 bx, by, bi;
  UINT32 scanx, scany;
  for (by = 0; by < downsampled_chroma_estimate.y.height / 8; by++)
  {
    for (bx = 0; bx < downsampled_chroma_estimate.y.width / 8; bx++)
    {
      if (possibility_count(refined_block_data.y, bx, by) > possibilities_limit)
      {
        // Copy the spatial domain information over the infeasible refined block.
        for (bi = 0; bi < 64; bi++)
        {
          scanx = bx * 8 + (bi % 8);
          scany = by * 8 + (bi / 8);
          refined_block_data.y.min[(by * refined_block_data.y.width / 8 + bx) * 64 + bi] = downsampled_chroma_estimate.y.min[scanx + scany * downsampled_chroma_estimate.y.width] - 128;
        }

        // DCT the block.
        ijg_fdct_islow(refined_block_data.y.min + ((by * refined_block_data.y.width / 8 + bx) * 64));

        // Quantize the results. For the IJG slow integer DCT, the quantization values are multiplied by eight.
        for (bi = 0; bi < 64; bi++)
        {
          refined_block_data.y.max[(by * refined_block_data.y.width / 8 + bx) * 64 + bi] = refined_block_data.y.min[(by * refined_block_data.y.width / 8 + bx) * 64 + bi] = quantize(refined_block_data.y.min[(by * refined_block_data.y.width / 8 + bx) * 64 + bi], (INT32)luma_table[bi]);
        }
      }
    }
  }

  for (by = 0; by < downsampled_chroma_estimate.cb.height / 8; by++)
  {
    for (bx = 0; bx < downsampled_chroma_estimate.cb.width / 8; bx++)
    {
      if (possibility_count(refined_block_data.cb, bx, by) > possibilities_limit)
      {
        // Copy the spatial domain information over the infeasible refined block.
        for (bi = 0; bi < 64; bi++)
        {
          scanx = bx * 8 + (bi % 8);
          scany = by * 8 + (bi / 8);
          refined_block_data.cb.min[(by * refined_block_data.cb.width / 8 + bx) * 64 + bi] = downsampled_chroma_estimate.cb.min[scanx + scany * downsampled_chroma_estimate.cb.width] - 128;
        }

        // DCT the block.
        ijg_fdct_islow(refined_block_data.cb.min + ((by * refined_block_data.cb.width / 8 + bx) * 64));

        // Quantize the results. For the IJG slow integer DCT, the quantization values are multiplied by eight.
        for (bi = 0; bi < 64; bi++)
        {
          refined_block_data.cb.max[(by * refined_block_data.cb.width / 8 + bx) * 64 + bi] = refined_block_data.cb.min[(by * refined_block_data.cb.width / 8 + bx) * 64 + bi] = quantize(refined_block_data.cb.min[(by * refined_block_data.cb.width / 8 + bx) * 64 + bi], (INT32)chroma_table[bi]);
        }
      }

      if (possibility_count(refined_block_data.cr, bx, by) > possibilities_limit)
      {
        // Copy the spatial domain information over the infeasible refined block.
        for (bi = 0; bi < 64; bi++)
        {
          scanx = bx * 8 + (bi % 8);
          scany = by * 8 + (bi / 8);
          refined_block_data.cr.min[(by * refined_block_data.cr.width / 8 + bx) * 64 + bi] = downsampled_chroma_estimate.cr.min[scanx + scany * downsampled_chroma_estimate.cr.width] - 128;
        }

        // DCT the block.
        ijg_fdct_islow(refined_block_data.cr.min + ((by * refined_block_data.cr.width / 8 + bx) * 64));

        // Quantize the results. For the IJG slow integer DCT, the quantization values are multiplied by eight.
        for (bi = 0; bi < 64; bi++)
        {
          refined_block_data.cr.max[(by * refined_block_data.cr.width / 8 + bx) * 64 + bi] = refined_block_data.cr.min[(by * refined_block_data.cr.width / 8 + bx) * 64 + bi] = quantize(refined_block_data.cr.min[(by * refined_block_data.cr.width / 8 + bx) * 64 + bi], (INT32)chroma_table[bi]);
        }
      }
    }
  }
}
