// Copyright (c) 2010 Andrew Lewis. All rights reserved.
// Use of this source code is governed the license in the LICENSE file.
//
// Please note: this file contains code equivalent to code in the Independent 
// JPEG Group JPEG decompressor version 6b:
//
//   http://www.ijg.org/
//

#include "fdctislow.h"

#define CONST_BITS  13
#define PASS1_BITS  2

#define DCTSIZE 8

#define FIX_0_298631336  ((INT32)  2446)	/* FIX(0.298631336) */
#define FIX_0_390180644  ((INT32)  3196)	/* FIX(0.390180644) */
#define FIX_0_541196100  ((INT32)  4433)	/* FIX(0.541196100) */
#define FIX_0_765366865  ((INT32)  6270)	/* FIX(0.765366865) */
#define FIX_0_899976223  ((INT32)  7373)	/* FIX(0.899976223) */
#define FIX_1_175875602  ((INT32)  9633)	/* FIX(1.175875602) */
#define FIX_1_501321110  ((INT32)  12299)	/* FIX(1.501321110) */
#define FIX_1_847759065  ((INT32)  15137)	/* FIX(1.847759065) */
#define FIX_1_961570560  ((INT32)  16069)	/* FIX(1.961570560) */
#define FIX_2_053119869  ((INT32)  16819)	/* FIX(2.053119869) */
#define FIX_2_562915447  ((INT32)  20995)	/* FIX(2.562915447) */
#define FIX_3_072711026  ((INT32)  25172)	/* FIX(3.072711026) */

#define MULTIPLY16C16(var,const) ((var) * (const))
#define MULTIPLY(var,const) MULTIPLY16C16(var,const)

#define RIGHT_SHIFT(x,shft) ((x) >> (shft))
#define ONE ((INT32) 1)
#define DESCALE(x,n)  RIGHT_SHIFT((x) + (ONE << ((n) - 1)), n)

// Perform a forward DCT in-place on the specific block.
void ijg_fdct_islow(INT16* data)
{
  INT32 tmp0, tmp1, tmp2, tmp3, tmp4, tmp5, tmp6, tmp7;
  INT32 tmp10, tmp11, tmp12, tmp13;
  INT32 z1, z2, z3, z4, z5;
  INT16* dataptr;
  int ctr;
  dataptr = data;
  for (ctr = DCTSIZE - 1; ctr >= 0; ctr--) 
  {
    tmp0 = dataptr[0] + dataptr[7];
    tmp7 = dataptr[0] - dataptr[7];
    tmp1 = dataptr[1] + dataptr[6];
    tmp6 = dataptr[1] - dataptr[6];
    tmp2 = dataptr[2] + dataptr[5];
    tmp5 = dataptr[2] - dataptr[5];
    tmp3 = dataptr[3] + dataptr[4];
    tmp4 = dataptr[3] - dataptr[4];

    tmp10 = tmp0 + tmp3;
    tmp13 = tmp0 - tmp3;
    tmp11 = tmp1 + tmp2;
    tmp12 = tmp1 - tmp2;
    
    dataptr[0] = (INT16)((tmp10 + tmp11) << PASS1_BITS);
    dataptr[4] = (INT16)((tmp10 - tmp11) << PASS1_BITS);
    
    z1 = MULTIPLY(tmp12 + tmp13, FIX_0_541196100);
    dataptr[2] = (INT16)DESCALE(z1 + MULTIPLY(tmp13, FIX_0_765366865), CONST_BITS - PASS1_BITS);
    dataptr[6] = (INT16)DESCALE(z1 + MULTIPLY(tmp12, - FIX_1_847759065), CONST_BITS - PASS1_BITS);
    
    z1 = tmp4 + tmp7;
    z2 = tmp5 + tmp6;
    z3 = tmp4 + tmp6;
    z4 = tmp5 + tmp7;
    z5 = MULTIPLY(z3 + z4, FIX_1_175875602);
    
    tmp4 = MULTIPLY(tmp4, FIX_0_298631336);
    tmp5 = MULTIPLY(tmp5, FIX_2_053119869);
    tmp6 = MULTIPLY(tmp6, FIX_3_072711026);
    tmp7 = MULTIPLY(tmp7, FIX_1_501321110);
    z1 = MULTIPLY(z1, - FIX_0_899976223);
    z2 = MULTIPLY(z2, - FIX_2_562915447);
    z3 = MULTIPLY(z3, - FIX_1_961570560);
    z4 = MULTIPLY(z4, - FIX_0_390180644);
    
    z3 += z5;
    z4 += z5;
    
    dataptr[7] = (INT16)DESCALE(tmp4 + z1 + z3, CONST_BITS - PASS1_BITS);
    dataptr[5] = (INT16)DESCALE(tmp5 + z2 + z4, CONST_BITS - PASS1_BITS);
    dataptr[3] = (INT16)DESCALE(tmp6 + z2 + z3, CONST_BITS - PASS1_BITS);
    dataptr[1] = (INT16)DESCALE(tmp7 + z1 + z4, CONST_BITS - PASS1_BITS);
    
    dataptr += DCTSIZE;
  }

  dataptr = data;
  for (ctr = DCTSIZE - 1; ctr >= 0; ctr--) 
  {
    tmp0 = dataptr[DCTSIZE * 0] + dataptr[DCTSIZE * 7];
    tmp7 = dataptr[DCTSIZE * 0] - dataptr[DCTSIZE * 7];
    tmp1 = dataptr[DCTSIZE * 1] + dataptr[DCTSIZE * 6];
    tmp6 = dataptr[DCTSIZE * 1] - dataptr[DCTSIZE * 6];
    tmp2 = dataptr[DCTSIZE * 2] + dataptr[DCTSIZE * 5];
    tmp5 = dataptr[DCTSIZE * 2] - dataptr[DCTSIZE * 5];
    tmp3 = dataptr[DCTSIZE * 3] + dataptr[DCTSIZE * 4];
    tmp4 = dataptr[DCTSIZE * 3] - dataptr[DCTSIZE * 4];
    
    tmp10 = tmp0 + tmp3;
    tmp13 = tmp0 - tmp3;
    tmp11 = tmp1 + tmp2;
    tmp12 = tmp1 - tmp2;
    
    dataptr[DCTSIZE * 0] = (INT16)DESCALE(tmp10 + tmp11, PASS1_BITS);
    dataptr[DCTSIZE * 4] = (INT16)DESCALE(tmp10 - tmp11, PASS1_BITS);
    
    z1 = MULTIPLY(tmp12 + tmp13, FIX_0_541196100);
    dataptr[DCTSIZE * 2] = (INT16)DESCALE(z1 + MULTIPLY(tmp13, FIX_0_765366865), CONST_BITS+PASS1_BITS);
    dataptr[DCTSIZE * 6] = (INT16)DESCALE(z1 + MULTIPLY(tmp12, - FIX_1_847759065), CONST_BITS+PASS1_BITS);
    
    z1 = tmp4 + tmp7;
    z2 = tmp5 + tmp6;
    z3 = tmp4 + tmp6;
    z4 = tmp5 + tmp7;
    z5 = MULTIPLY(z3 + z4, FIX_1_175875602);
    
    tmp4 = MULTIPLY(tmp4, FIX_0_298631336);
    tmp5 = MULTIPLY(tmp5, FIX_2_053119869);
    tmp6 = MULTIPLY(tmp6, FIX_3_072711026);
    tmp7 = MULTIPLY(tmp7, FIX_1_501321110);
    z1 = MULTIPLY(z1, - FIX_0_899976223);
    z2 = MULTIPLY(z2, - FIX_2_562915447);
    z3 = MULTIPLY(z3, - FIX_1_961570560);
    z4 = MULTIPLY(z4, - FIX_0_390180644);
    
    z3 += z5;
    z4 += z5;
    
    dataptr[DCTSIZE * 7] = (INT16) DESCALE(tmp4 + z1 + z3, CONST_BITS + PASS1_BITS);
    dataptr[DCTSIZE * 5] = (INT16) DESCALE(tmp5 + z2 + z4, CONST_BITS + PASS1_BITS);
    dataptr[DCTSIZE * 3] = (INT16) DESCALE(tmp6 + z2 + z3, CONST_BITS + PASS1_BITS);
    dataptr[DCTSIZE * 1] = (INT16) DESCALE(tmp7 + z1 + z4, CONST_BITS + PASS1_BITS);
    
    dataptr++;
  }
}
