// Copyright (c) 2010 Andrew Lewis. All rights reserved.
// Use of this source code is governed the license in the LICENSE file.
//
// Please note: this file contains code equivalent to code in the Independent 
// JPEG Group JPEG decompressor version 6b:
//
//   http://www.ijg.org/
//

#include "forwardoperations.h"

void block_idct(INT16* input)
{
  UINT8 ctr;
  INT32 tmp0, tmp1, tmp2, tmp3;
  INT32 tmp10, tmp11, tmp12, tmp13;
  INT32 z1, z2, z3, z4, z5;
  INT32 workspace[64];
  INT32* wsptr;
  INT16* inptr;
  INT16* outptr;
  
  inptr = input;
  wsptr = workspace;
  for (ctr = 8; ctr > 0; ctr--) 
  {
    if (inptr[8 * 1] == 0 && inptr[8 * 2] == 0 &&
	inptr[8 * 3] == 0 && inptr[8 * 4] == 0 &&
	inptr[8 * 5] == 0 && inptr[8 * 6] == 0 &&
	inptr[8 * 7] == 0) 
    {
      int dcval = inptr[8 * 0] << PASS1_BITS;
      
      wsptr[8 * 0] = dcval;
      wsptr[8 * 1] = dcval;
      wsptr[8 * 2] = dcval;
      wsptr[8 * 3] = dcval;
      wsptr[8 * 4] = dcval;
      wsptr[8 * 5] = dcval;
      wsptr[8 * 6] = dcval;
      wsptr[8 * 7] = dcval;
      
      inptr++;
      wsptr++;
      continue;
    }
    
    z2 = inptr[8 * 2];
    z3 = inptr[8 * 6];
    
    z1 = MULTIPLY(z2 + z3, FIX_0_541196100);
    tmp2 = z1 + MULTIPLY(z3, - FIX_1_847759065);
    tmp3 = z1 + MULTIPLY(z2, FIX_0_765366865);
    
    z2 = inptr[8 * 0];
    z3 = inptr[8 * 4];

    tmp0 = (z2 + z3) << CONST_BITS;
    tmp1 = (z2 - z3) << CONST_BITS;
    
    tmp10 = tmp0 + tmp3;
    tmp13 = tmp0 - tmp3;
    tmp11 = tmp1 + tmp2;
    tmp12 = tmp1 - tmp2;
    
    tmp0 = inptr[8 * 7];
    tmp1 = inptr[8 * 5];
    tmp2 = inptr[8 * 3];
    tmp3 = inptr[8 * 1];
    
    z1 = tmp0 + tmp3;
    z2 = tmp1 + tmp2;
    z3 = tmp0 + tmp2;
    z4 = tmp1 + tmp3;
    z5 = MULTIPLY(z3 + z4, FIX_1_175875602);
    
    tmp0 = MULTIPLY(tmp0, FIX_0_298631336);
    tmp1 = MULTIPLY(tmp1, FIX_2_053119869);
    tmp2 = MULTIPLY(tmp2, FIX_3_072711026);
    tmp3 = MULTIPLY(tmp3, FIX_1_501321110);
    z1 = MULTIPLY(z1, - FIX_0_899976223);
    z2 = MULTIPLY(z2, - FIX_2_562915447);
    z3 = MULTIPLY(z3, - FIX_1_961570560);
    z4 = MULTIPLY(z4, - FIX_0_390180644);
    
    z3 += z5;
    z4 += z5;
    
    tmp0 += z1 + z3;
    tmp1 += z2 + z4;
    tmp2 += z2 + z3;
    tmp3 += z1 + z4;
    
    wsptr[8 * 0] = (int) DESCALE(tmp10 + tmp3, CONST_BITS - PASS1_BITS);
    wsptr[8 * 7] = (int) DESCALE(tmp10 - tmp3, CONST_BITS - PASS1_BITS);
    wsptr[8 * 1] = (int) DESCALE(tmp11 + tmp2, CONST_BITS - PASS1_BITS);
    wsptr[8 * 6] = (int) DESCALE(tmp11 - tmp2, CONST_BITS - PASS1_BITS);
    wsptr[8 * 2] = (int) DESCALE(tmp12 + tmp1, CONST_BITS - PASS1_BITS);
    wsptr[8 * 5] = (int) DESCALE(tmp12 - tmp1, CONST_BITS - PASS1_BITS);
    wsptr[8 * 3] = (int) DESCALE(tmp13 + tmp0, CONST_BITS - PASS1_BITS);
    wsptr[8 * 4] = (int) DESCALE(tmp13 - tmp0, CONST_BITS - PASS1_BITS);
    
    inptr++;
    wsptr++;
  }

  wsptr = workspace;
  for (ctr = 0; ctr < 8; ctr++) {
    outptr = input + (8 * ctr);
    
#ifndef NO_ZERO_ROW_TEST
    if (wsptr[1] == 0 && wsptr[2] == 0 && wsptr[3] == 0 && wsptr[4] == 0 &&
	wsptr[5] == 0 && wsptr[6] == 0 && wsptr[7] == 0) {
      INT16 dcval = RANGE_LIMIT((int) DESCALE((INT32) wsptr[0], PASS1_BITS+3));
      
      outptr[0] = dcval;
      outptr[1] = dcval;
      outptr[2] = dcval;
      outptr[3] = dcval;
      outptr[4] = dcval;
      outptr[5] = dcval;
      outptr[6] = dcval;
      outptr[7] = dcval;

      wsptr += 8;
      continue;
    }
#endif
    
    z2 = (INT32) wsptr[2];
    z3 = (INT32) wsptr[6];
    
    z1 = MULTIPLY(z2 + z3, FIX_0_541196100);
    tmp2 = z1 + MULTIPLY(z3, - FIX_1_847759065);
    tmp3 = z1 + MULTIPLY(z2, FIX_0_765366865);
    
    tmp0 = ((INT32) wsptr[0] + (INT32) wsptr[4]) << CONST_BITS;
    tmp1 = ((INT32) wsptr[0] - (INT32) wsptr[4]) << CONST_BITS;
    
    tmp10 = tmp0 + tmp3;
    tmp13 = tmp0 - tmp3;
    tmp11 = tmp1 + tmp2;
    tmp12 = tmp1 - tmp2;
    
    tmp0 = (INT32) wsptr[7];
    tmp1 = (INT32) wsptr[5];
    tmp2 = (INT32) wsptr[3];
    tmp3 = (INT32) wsptr[1];
    
    z1 = tmp0 + tmp3;
    z2 = tmp1 + tmp2;
    z3 = tmp0 + tmp2;
    z4 = tmp1 + tmp3;
    z5 = MULTIPLY(z3 + z4, FIX_1_175875602);
    
    tmp0 = MULTIPLY(tmp0, FIX_0_298631336);
    tmp1 = MULTIPLY(tmp1, FIX_2_053119869);
    tmp2 = MULTIPLY(tmp2, FIX_3_072711026);
    tmp3 = MULTIPLY(tmp3, FIX_1_501321110);
    z1 = MULTIPLY(z1, - FIX_0_899976223);
    z2 = MULTIPLY(z2, - FIX_2_562915447);
    z3 = MULTIPLY(z3, - FIX_1_961570560);
    z4 = MULTIPLY(z4, - FIX_0_390180644);
    
    z3 += z5;
    z4 += z5;
    
    tmp0 += z1 + z3;
    tmp1 += z2 + z4;
    tmp2 += z2 + z3;
    tmp3 += z1 + z4;
    
    outptr[0] = RANGE_LIMIT((int) DESCALE(tmp10 + tmp3, CONST_BITS + PASS1_BITS + 3));
    outptr[7] = RANGE_LIMIT((int) DESCALE(tmp10 - tmp3, CONST_BITS + PASS1_BITS + 3));
    outptr[1] = RANGE_LIMIT((int) DESCALE(tmp11 + tmp2, CONST_BITS + PASS1_BITS + 3));
    outptr[6] = RANGE_LIMIT((int) DESCALE(tmp11 - tmp2, CONST_BITS + PASS1_BITS + 3));
    outptr[2] = RANGE_LIMIT((int) DESCALE(tmp12 + tmp1, CONST_BITS + PASS1_BITS + 3));
    outptr[5] = RANGE_LIMIT((int) DESCALE(tmp12 - tmp1, CONST_BITS + PASS1_BITS + 3));
    outptr[3] = RANGE_LIMIT((int) DESCALE(tmp13 + tmp0, CONST_BITS + PASS1_BITS + 3));
    outptr[4] = RANGE_LIMIT((int) DESCALE(tmp13 - tmp0, CONST_BITS + PASS1_BITS + 3));
    
    wsptr += 8;
  }
}

UINT64 possibility_count(const interval_plane block_plane, const UINT8 bx, const UINT8 by)
{
  UINT8 bi;
  UINT64 result = 1L;
  UINT64 old_result = 1L;

  // TODO: Saturate at overflow.

  // Work out the possibilities using the size of the range on each quantized DCT coefficient.
  for (bi = 0; bi < 64; bi++)
  {
    result *= (block_plane.max[((by * block_plane.width / 8) + bx) * 64 + bi] - block_plane.min[((by * block_plane.width / 8) + bx) * 64 + bi]) + 1;
    if (result < old_result)
      return (1L << 63);
    old_result = result;
  }

  return result;
}

interval_plane refine_dct_plane(interval_plane block_plane, interval_plane refined_block_plane, quantization_table q_table, interval_plane current_spatial_plane, block_status_summary* block_status, FILE* additional_output_file)
{
  INT16* working = (INT16*)malloc(64 * sizeof(INT16));
  INT16* working_scratch = (INT16*)malloc(64 * sizeof(INT16));
  UINT8 bx, by, bi;
  boolean had_success, first_duplicate;
  interval_plane result_plane = create_interval_plane(current_spatial_plane.width, current_spatial_plane.height);
  block_status->infeasible_block_count = block_status->ambiguous_block_count = block_status->impossible_block_count = block_status->exact_block_count = 0;

  // Search through the possibilities, dequantizing and performing a DCT at each.
  for (by = 0; by < block_plane.height / 8; by++)
  {
    for (bx = 0; bx < block_plane.width / 8; bx++)
    {
      // Check whether it's feasible to search this block.
      if (possibility_count(block_plane, bx, by) > possibilities_limit)
      {
        // Copy the input to the output.
        for (bi = 0; bi < 64; bi++)
        {
          result_plane.min[((by * block_plane.width * 8) + bx * 8) + (bi / 8) * block_plane.width + (bi % 8)] = current_spatial_plane.min[((by * block_plane.width * 8) + bx * 8) + (bi / 8) * block_plane.width + (bi % 8)];
          result_plane.max[((by * block_plane.width * 8) + bx * 8) + (bi / 8) * block_plane.width + (bi % 8)] = current_spatial_plane.max[((by * block_plane.width * 8) + bx * 8) + (bi / 8) * block_plane.width + (bi % 8)];

          refined_block_plane.min[((by * refined_block_plane.width / 8) + bx) * 64 + bi] = block_plane.min[((by * block_plane.width / 8) + bx) * 64 + bi];
          refined_block_plane.max[((by * refined_block_plane.width / 8) + bx) * 64 + bi] = block_plane.max[((by * block_plane.width / 8) + bx) * 64 + bi];
        }

        // If additional output was requested, write this data out to the additional output file.
        if (additional_output_file != NULL)
        {
          fprintf(additional_output_file, "Infeasible (%d, %d):\n", bx, by);
          for (bi = 0; bi < 64; bi++)
          {
            if (bi % 8 == 0)
              fprintf(additional_output_file, "\n");
            fprintf(additional_output_file, "(%d, %d) ", block_plane.min[((by * block_plane.width / 8) + bx) * 64 + bi], block_plane.max[((by * block_plane.width / 8) + bx) * 64 + bi]);
          }
          fprintf(additional_output_file, "\n");
        }

        block_status->infeasible_block_count++;

        continue;
      }

      for (bi = 0; bi < 64; bi++)
        working[bi] = (INT16)block_plane.min[((by * block_plane.width / 8) + bx) * 64 + bi];

      // We loop until the final for loop here exits, indicating that there are no more valid possibilities for the block.
      // working stores the current quantized DCT coefficient choices, incremented each time to check all possibilities.
      // working_scratch stores the unquantized DCT coefficients for working, and then has a IDCT performed in-place.
      // result_plane stores the intervals for the result of the IDCT.
      bi = 0;
      had_success = FALSE;
      first_duplicate = FALSE;
      while (bi != 64)
      {
        // Test the current block.
        for (bi = 0; bi < 64; bi++)
          working_scratch[bi] = (working[bi] * (INT16)q_table[bi]);

        block_idct(working_scratch);

        // Work out whether this result is consistent with our previous estimates.
        for (bi = 0; bi < 64; bi++)
        {
          if (working_scratch[bi] > current_spatial_plane.max[((by * block_plane.width * 8) + bx * 8) + (bi / 8) * block_plane.width + (bi % 8)] || 
              working_scratch[bi] < current_spatial_plane.min[((by * block_plane.width * 8) + bx * 8) + (bi / 8) * block_plane.width + (bi % 8)])
            break;
        }

        if (bi == 64)
        {
          // If this is the first working block, set that value. Otherwise, union it with the old value.
          if (had_success)
          {
            for (bi = 0; bi < 64; bi++)
            {
              if (working_scratch[bi] < result_plane.min[((by * block_plane.width * 8) + bx * 8) + (bi / 8) * block_plane.width + (bi % 8)])
                result_plane.min[((by * block_plane.width * 8) + bx * 8) + (bi / 8) * block_plane.width + (bi % 8)] = working_scratch[bi];
              if (working_scratch[bi] > result_plane.max[((by * block_plane.width * 8) + bx * 8) + (bi / 8) * block_plane.width + (bi % 8)])
                result_plane.max[((by * block_plane.width * 8) + bx * 8) + (bi / 8) * block_plane.width + (bi % 8)] = working_scratch[bi];
  
              if (working[bi] < refined_block_plane.min[((by * refined_block_plane.width / 8) + bx) * 64 + bi])
                refined_block_plane.min[((by * refined_block_plane.width / 8) + bx) * 64 + bi] = working[bi];
              if (working[bi] > refined_block_plane.max[((by * refined_block_plane.width / 8) + bx) * 64 + bi])
                refined_block_plane.max[((by * refined_block_plane.width / 8) + bx) * 64 + bi] = working[bi];
            }
  
            if (first_duplicate)
            {
              block_status->ambiguous_block_count++;
              block_status->exact_block_count--;

              first_duplicate = FALSE;
            }
          }
          else
          {
            for (bi = 0; bi < 64; bi++)
            {
              result_plane.min[((by * block_plane.width * 8) + bx * 8) + (bi / 8) * block_plane.width + (bi % 8)] = working_scratch[bi];
              result_plane.max[((by * block_plane.width * 8) + bx * 8) + (bi / 8) * block_plane.width + (bi % 8)] = working_scratch[bi];

              refined_block_plane.min[((by * refined_block_plane.width / 8) + bx) * 64 + bi] = working[bi];
              refined_block_plane.max[((by * refined_block_plane.width / 8) + bx) * 64 + bi] = working[bi];
            }

            block_status->exact_block_count++;

            // If there is another possibility, it's the first duplicate.
            first_duplicate = TRUE;
          }
          had_success = TRUE;
        }

        // Increment to the next possible block.
        for (bi = 0; bi < 64; bi++)
        {
          if (working[bi] < (INT16)block_plane.max[((by * block_plane.width / 8) + bx) * 64 + bi])
          {
            // Increment and test this one.
            working[bi]++;
            break;
          }
          else if (working[bi] == (INT16)block_plane.max[((by * block_plane.width / 8) + bx) * 64 + bi])
          {
            // Set this one back to zero and continue to carry.
            working[bi] = (INT16)block_plane.min[((by * block_plane.width / 8) + bx) * 64 + bi];
          }
        }
      }

      if (!had_success)
        block_status->impossible_block_count++;
    }
  }

  free(working);
  free(working_scratch);

  return result_plane;
}

void full_decompression_refinement(block_interval_component_image_data refined_values, quantization_table luma_table, quantization_table chroma_table, computation_plane smoothing_computations, inverse_ycc_rgb_table cspace_tables, packed_image_data image, block_status_summary* y_block_status, block_status_summary* cb_block_status, block_status_summary* cr_block_status, FILE* additional_output_file)
{
  UINT8 bx, by, bi, bi_cb, bi_cr, bi_y, y_element;
  boolean check_cb, check_cr, single_cb, single_cr, check_y1, check_y2, check_y3, check_y4, single_y, entire_block_correct, had_solution;
  INT16* working_cb = (INT16*)malloc(64 * sizeof(INT16));
  INT16* working_cr = (INT16*)malloc(64 * sizeof(INT16));
  INT16* working_y1 = (INT16*)malloc(64 * sizeof(INT16));
  INT16* working_y2 = (INT16*)malloc(64 * sizeof(INT16));
  INT16* working_y3 = (INT16*)malloc(64 * sizeof(INT16));
  INT16* working_y4 = (INT16*)malloc(64 * sizeof(INT16));
  INT16* working_y = (INT16*)malloc(64 * sizeof(INT16));
  INT16* working_scratch_y = (INT16*)malloc(64 * sizeof(INT16));
  INT16* working_scratch_cb = (INT16*)malloc(64 * sizeof(INT16));
  INT16* working_scratch_cr = (INT16*)malloc(64 * sizeof(INT16));
  INT32* smoothed_block = (INT32*)malloc(256 * sizeof(INT32));
  INT16* working_cb_block_min = (INT16*)malloc(64 * sizeof(INT16));
  INT16* working_cb_block_max = (INT16*)malloc(64 * sizeof(INT16));
  INT16* working_cr_block_min = (INT16*)malloc(64 * sizeof(INT16));
  INT16* working_cr_block_max = (INT16*)malloc(64 * sizeof(INT16));
  INT16* working_y_block_min = (INT16*)malloc(64 * sizeof(INT16));
  INT16* working_y_block_max = (INT16*)malloc(64 * sizeof(INT16));
  UINT16 y_block_start = 0;
  interval_component_image_data downsampled_spatial_data = create_interval_component_image_data(image, TRUE);
  long int cb_min, cb_max, cr_min, cr_max, yl_min, yl_max;
  INT32 r_full, g_full, b_full;
  UINT8 r, g, b;
  UINT16 image_x, image_y;

  // Loop over the chroma blocks to find a currently ambiguous block.
  for (by = 0; by < refined_values.cb.height / 8; by++)
  {
    for (bx = 0; bx < refined_values.cb.width / 8; bx++)
    {
      check_cb = (possibility_count(refined_values.cb, bx, by) != 1L && possibility_count(refined_values.cb, bx, by) <= possibilities_limit);
      check_cr = (possibility_count(refined_values.cr, bx, by) != 1L && possibility_count(refined_values.cr, bx, by) <= possibilities_limit);
      single_y = (possibility_count(refined_values.y, bx * 2, by * 2) == 1L && possibility_count(refined_values.y, bx * 2 + 1, by * 2) == 1L && possibility_count(refined_values.y, bx * 2, by * 2 + 1) == 1L && possibility_count(refined_values.y, bx * 2 + 1, by * 2 + 1) == 1L);
      single_cb = (possibility_count(refined_values.cb, bx, by) == 1L);
      single_cr = (possibility_count(refined_values.cr, bx, by) == 1L);
      check_y1 = (possibility_count(refined_values.y, bx * 2, by * 2) != 1L && possibility_count(refined_values.y, bx * 2, by * 2) < possibilities_limit);
      check_y2 = (possibility_count(refined_values.y, bx * 2 + 1, by * 2) != 1L && possibility_count(refined_values.y, bx * 2 + 1, by * 2) < possibilities_limit);
      check_y3 = (possibility_count(refined_values.y, bx * 2, by * 2 + 1) != 1L && possibility_count(refined_values.y, bx * 2, by * 2 + 1) < possibilities_limit);
      check_y4 = (possibility_count(refined_values.y, bx * 2 + 1, by * 2 + 1) != 1L && possibility_count(refined_values.y, bx * 2 + 1, by * 2 + 1) < possibilities_limit);

      // First for chroma. Non-ambiguous blocks don't need any checking.
      if ((check_cb || check_cr) && single_y && possibility_count(refined_values.cb, bx, by) <= possibilities_limit && possibility_count(refined_values.cr, bx, by) <= possibilities_limit && possibility_count(refined_values.cb, bx, by) * possibility_count(refined_values.cr, bx, by) <= possibilities_limit)
      {
        // Copy the quantized values.
        for (bi = 0; bi < 64; bi++)
        {
          working_cb[bi] = refined_values.cb.min[(by * refined_values.cb.width / 8 + bx) * 64 + bi];
          working_cr[bi] = refined_values.cr.min[(by * refined_values.cr.width / 8 + bx) * 64 + bi];

          working_y1[bi] = refined_values.y.max[((by * 2) * refined_values.y.width / 8 + (bx * 2)) * 64 + bi] * (INT16)luma_table[bi];
          working_y2[bi] = refined_values.y.max[((by * 2) * refined_values.y.width / 8 + (bx * 2) + 1) * 64 + bi] * (INT16)luma_table[bi];
          working_y3[bi] = refined_values.y.max[(((by * 2) + 1) * refined_values.y.width / 8 + (bx * 2)) * 64 + bi] * (INT16)luma_table[bi];
          working_y4[bi] = refined_values.y.max[(((by * 2) + 1) * refined_values.y.width / 8 + (bx * 2) + 1) * 64 + bi] * (INT16)luma_table[bi];
        }

        // IDCT the luma data.
        block_idct(working_y1);
        block_idct(working_y2);
        block_idct(working_y3);
        block_idct(working_y4);

        had_solution = FALSE;

        // Work through the possibilities.
        bi_cb = bi_cr = 0;
        while (bi_cr != 64)
        {
          // Check the current value.
          for (bi = 0; bi < 64; bi++)
          {
            working_scratch_cr[bi] = (working_cr[bi] * (INT16)chroma_table[bi]);
            working_scratch_cb[bi] = (working_cb[bi] * (INT16)chroma_table[bi]);
          }
  
          block_idct(working_scratch_cr);
          block_idct(working_scratch_cb);
  
          // Complete the decompression process on this block, to check whether the possibility is valid.

          // Copy the chroma into the downsampled_spatial_data.
          for (bi = 0; bi < 64; bi++)
          {
            downsampled_spatial_data.cb.min[(by * 8 + bi / 8) * downsampled_spatial_data.cb.width + bx * 8 + bi % 8] = downsampled_spatial_data.cb.max[(by * 8 + bi / 8) * downsampled_spatial_data.cb.width + bx * 8 + bi % 8] = working_scratch_cb[bi];
            downsampled_spatial_data.cr.min[(by * 8 + bi / 8) * downsampled_spatial_data.cr.width + bx * 8 + bi % 8] = downsampled_spatial_data.cr.max[(by * 8 + bi / 8) * downsampled_spatial_data.cr.width + bx * 8 + bi % 8] = working_scratch_cr[bi];
          }

          // Smooth the chroma and check the resulting sample values.
          entire_block_correct = TRUE;
          for (image_y = by * 16 + 1; image_y < (by + 1) * 16 - 1; image_y++)
          {
            for (image_x = bx * 16 + 1; image_x < (bx + 1) * 16 - 1; image_x++)
            {
              // Get the component values corresponding to (image_x, image_y) in the output image.
              yl_min = yl_max = (image_y - by * 16) < 8 ? 
                (image_x - bx * 16 < 8 ? 
                  (working_y1[(image_y % 8) * 8 + image_x % 8]) : 
                  (working_y2[(image_y % 8) * 8 + image_x % 8])) : 
                (image_x - bx * 16 < 8 ? 
                  (working_y3[(image_y % 8) * 8 + image_x % 8]) : 
                  (working_y4[(image_y % 8) * 8 + image_x % 8]));
              evaluate_computation(smoothing_computations.computations[image_y * smoothing_computations.width + image_x], &cb_max, &cb_min, downsampled_spatial_data.cb);
              evaluate_computation(smoothing_computations.computations[image_y * smoothing_computations.width + image_x], &cr_max, &cr_min, downsampled_spatial_data.cr);

              // Perform colour space conversion.
              r_full = yl_min + cspace_tables.cr_r_tab[(UINT8)cr_min];
              g_full = yl_min + ((INT16)RIGHT_SHIFT(cspace_tables.cb_g_tab[cb_min] + cspace_tables.cr_g_tab[cr_min], SCALEBITS));
              b_full = yl_min + cspace_tables.cb_b_tab[cb_min];
              r = r_full > MAX_JSAMPLE ? MAX_JSAMPLE : (r_full < 0 ? 0 : r_full);
              g = g_full > MAX_JSAMPLE ? MAX_JSAMPLE : (g_full < 0 ? 0 : g_full);
              b = b_full > MAX_JSAMPLE ? MAX_JSAMPLE : (b_full < 0 ? 0 : b_full);

              if (image.samples[3 * (image_y * image.width + image_x)] != r ||
                image.samples[3 * (image_y * image.width + image_x) + 1] != g ||
                image.samples[3 * (image_y * image.width + image_x) + 2] != b)
              {
                entire_block_correct = FALSE;
                break;
              }
            }
            if (!entire_block_correct)
              break;
          }

          // If this pair of Cb, Cr block possibilities worked, union it with the correct set.
          if (entire_block_correct)
          {
            if (!had_solution)
            {
              // First solution.
              for (bi = 0; bi < 64; bi++)
              {
                working_cb_block_min[bi] = working_cb_block_max[bi] = working_cb[bi];
                working_cr_block_min[bi] = working_cr_block_max[bi] = working_cr[bi];
              }
              had_solution = TRUE;
            }
            else
            {
              // Union with previous solution.
              for (bi = 0; bi < 64; bi++)
              {
                if (working_cb[bi] < working_cb_block_min[bi])
                  working_cb_block_min[bi] = working_cb[bi];
                if (working_cb[bi] > working_cb_block_max[bi])
                  working_cb_block_max[bi] = working_cb[bi];
                if (working_cr[bi] < working_cr_block_min[bi])
                  working_cr_block_min[bi] = working_cr[bi];
                if (working_cr[bi] > working_cr_block_max[bi])
                  working_cr_block_max[bi] = working_cr[bi];
              }

              if (additional_output_file != NULL)
              {
                fprintf(additional_output_file, "(Cb/Cr) Multiple full decompression solutions at (%d, %d)\n", bx, by);
                fprintf(additional_output_file, "Cb\n");
                for (bi = 0; bi < 64; bi++)
                {
                  if (bi % 8 == 0)
                    fprintf(additional_output_file, "\n");
                  fprintf(additional_output_file, "%d ", working_cb[bi]);
                }
                fprintf(additional_output_file, "\n");
                fprintf(additional_output_file, "Cr\n");
                for (bi = 0; bi < 64; bi++)
                {
                  if (bi % 8 == 0)
                    fprintf(additional_output_file, "\n");
                  fprintf(additional_output_file, "%d ", working_cr[bi]);
                }
                fprintf(additional_output_file, "\n");
              }
            }
          }
  
          // Increment to the next possible block.
          for (bi_cb = 0; bi_cb < 64; bi_cb++)
          {
            if (working_cb[bi_cb] < (INT16)refined_values.cb.max[((by * refined_values.cb.width / 8) + bx) * 64 + bi_cb])
            {
              // Increment and test this one.
              working_cb[bi_cb]++;
              break;
            }
            else if (working_cb[bi_cb] == (INT16)refined_values.cb.max[((by * refined_values.cb.width / 8) + bx) * 64 + bi_cb])
            {
              // Set this one back to zero and continue to carry.
              working_cb[bi_cb] = (INT16)refined_values.cb.min[((by * refined_values.cb.width / 8) + bx) * 64 + bi_cb];
            }
          }

          if (bi_cb == 64)
          {
            for (bi_cr = 0; bi_cr < 64; bi_cr++)
            {
              if (working_cr[bi_cr] < (INT16)refined_values.cr.max[((by * refined_values.cr.width / 8) + bx) * 64 + bi_cr])
              {
                // Increment and test this one.
                working_cr[bi_cr]++;
                break;
              }
              else if (working_cr[bi_cr] == (INT16)refined_values.cr.max[((by * refined_values.cr.width / 8) + bx) * 64 + bi_cr])
              {
                // Set this one back to zero and continue to carry.
                working_cr[bi_cr] = (INT16)refined_values.cr.min[((by * refined_values.cr.width / 8) + bx) * 64 + bi_cr];
              }
            }
          }
          else
            bi_cr = 0;
        }

        // Copy the hopefully refined sets back into the output data.
        for (bi = 0; bi < 64; bi++)
        {
//          printf("Updating %d %d Cr %d %d to ", refined_values.cb.min[(by * refined_values.cb.width / 8 + bx) * 64 + bi], refined_values.cb.max[(by * refined_values.cb.width / 8 + bx) * 64 + bi], refined_values.cr.min[(by * refined_values.cr.width / 8 + bx) * 64 + bi], refined_values.cr.max[(by * refined_values.cr.width / 8 + bx) * 64 + bi]);

          refined_values.cb.min[(by * refined_values.cb.width / 8 + bx) * 64 + bi] = working_cb_block_min[bi];
          refined_values.cb.max[(by * refined_values.cb.width / 8 + bx) * 64 + bi] = working_cb_block_max[bi];
          refined_values.cr.min[(by * refined_values.cr.width / 8 + bx) * 64 + bi] = working_cr_block_min[bi];
          refined_values.cr.max[(by * refined_values.cr.width / 8 + bx) * 64 + bi] = working_cr_block_max[bi];

//          printf("%d %d Cr %d %d\n", refined_values.cb.min[(by * refined_values.cb.width / 8 + bx) * 64 + bi], refined_values.cb.max[(by * refined_values.cb.width / 8 + bx) * 64 + bi], refined_values.cr.min[(by * refined_values.cr.width / 8 + bx) * 64 + bi], refined_values.cr.max[(by * refined_values.cr.width / 8 + bx) * 64 + bi]);
        }

        if (check_cb && possibility_count(refined_values.cb, bx, by) == 1L)
        {
          cb_block_status->ambiguous_block_count--;
          cb_block_status->exact_block_count++;
        }
        
        if (check_cr && possibility_count(refined_values.cr, bx, by) == 1L)
        {
          cr_block_status->ambiguous_block_count--;
          cr_block_status->exact_block_count++;
        }
      }

      if ((check_y1 || check_y2 || check_y3 || check_y4) && single_cb && single_cr)
      {
        for (y_element = 0; y_element < 4; y_element++)
        {
          if ((y_element == 0 && !check_y1) || (y_element == 1 && !check_y2) || (y_element == 2 && !check_y3) || (y_element == 3 && !check_y4))
            continue;

          if (check_y1 && y_element == 0)
            y_block_start = ((by * 2) * refined_values.y.width / 8 + (bx * 2)) * 64;
          if (check_y2 && y_element == 1)
            y_block_start = ((by * 2) * refined_values.y.width / 8 + (bx * 2) + 1) * 64;
          if (check_y3 && y_element == 2)
            y_block_start = (((by * 2) + 1) * refined_values.y.width / 8 + (bx * 2)) * 64;
          if (check_y4 && y_element == 3)
            y_block_start = (((by * 2) + 1) * refined_values.y.width / 8 + (bx * 2) + 1) * 64;
  
          // Copy the quantized values.
          for (bi = 0; bi < 64; bi++)
          {
            working_cb[bi] = refined_values.cb.min[(by * refined_values.cb.width / 8 + bx) * 64 + bi] * (INT16)chroma_table[bi];
            working_cr[bi] = refined_values.cr.min[(by * refined_values.cr.width / 8 + bx) * 64 + bi] * (INT16)chroma_table[bi];
  
            working_y[bi] = refined_values.y.min[y_block_start + bi];
          }
  
          // IDCT the chroma data.
          block_idct(working_cb);
          block_idct(working_cr);
  
          had_solution = FALSE;
  
          // Work through the possibilities.
          bi_y = 0;
          while (bi_y != 64)
          {
            // Check the current value.
            for (bi = 0; bi < 64; bi++)
            {
              working_scratch_y[bi] = (working_y[bi] * (INT16)luma_table[bi]);
            }
    
            block_idct(working_scratch_y);
    
            // Complete the decompression process on this block, to check whether the possibility is valid.
  
            // Copy the chroma into the downsampled_spatial_data.
            for (bi = 0; bi < 64; bi++)
            {
              downsampled_spatial_data.cb.min[(by * 8 + bi / 8) * downsampled_spatial_data.cb.width + bx * 8 + bi % 8] = downsampled_spatial_data.cb.max[(by * 8 + bi / 8) * downsampled_spatial_data.cb.width + bx * 8 + bi % 8] = working_cb[bi];
              downsampled_spatial_data.cr.min[(by * 8 + bi / 8) * downsampled_spatial_data.cr.width + bx * 8 + bi % 8] = downsampled_spatial_data.cr.max[(by * 8 + bi / 8) * downsampled_spatial_data.cr.width + bx * 8 + bi % 8] = working_cr[bi];
            }
  
            // Smooth the chroma and check the resulting sample values.
            entire_block_correct = TRUE;
            for (image_y = ((y_element == 0 || y_element == 1) ? by * 16 + 1 : by * 16 + 9); image_y < ((y_element == 0 || y_element == 1) ? by * 16 + 7 : (by + 1) * 16 - 1); image_y++)
            {
              for (image_x = ((y_element == 0 || y_element == 2) ? bx * 16 + 1 : bx * 16 + 9); image_x < ((y_element == 0 || y_element == 2) ? bx * 16 + 7 : (bx + 1) * 16 - 1); image_x++)
              {
                // Get the component values corresponding to (image_x, image_y) in the output image.
                yl_min = yl_max = working_scratch_y[(image_y % 8) * 8 + image_x % 8];
                evaluate_computation(smoothing_computations.computations[image_y * smoothing_computations.width + image_x], &cb_max, &cb_min, downsampled_spatial_data.cb);
                evaluate_computation(smoothing_computations.computations[image_y * smoothing_computations.width + image_x], &cr_max, &cr_min, downsampled_spatial_data.cr);
  
                // Perform colour space conversion.
                r_full = yl_min + cspace_tables.cr_r_tab[(UINT8)cr_min];
                g_full = yl_min + ((INT16)RIGHT_SHIFT(cspace_tables.cb_g_tab[cb_min] + cspace_tables.cr_g_tab[cr_min], SCALEBITS));
                b_full = yl_min + cspace_tables.cb_b_tab[cb_min];
                r = r_full > MAX_JSAMPLE ? MAX_JSAMPLE : (r_full < 0 ? 0 : r_full);
                g = g_full > MAX_JSAMPLE ? MAX_JSAMPLE : (g_full < 0 ? 0 : g_full);
                b = b_full > MAX_JSAMPLE ? MAX_JSAMPLE : (b_full < 0 ? 0 : b_full);
  
                if (image.samples[3 * (image_y * image.width + image_x)] != r ||
                  image.samples[3 * (image_y * image.width + image_x) + 1] != g ||
                  image.samples[3 * (image_y * image.width + image_x) + 2] != b)
                {
                  entire_block_correct = FALSE;
                  break;
                }
              }
              if (!entire_block_correct)
                break;
            }
  
            // If this Y block possibility worked, union it with the correct set.
            if (entire_block_correct)
            {
              if (!had_solution)
              {
                // First solution.
                for (bi = 0; bi < 64; bi++)
                {
                  working_y_block_min[bi] = working_y_block_max[bi] = working_y[bi];
                }
                had_solution = TRUE;
              }
              else
              {
                // Union with previous solution.
                for (bi = 0; bi < 64; bi++)
                {
                  if (working_y[bi] < working_y_block_min[bi])
                    working_y_block_min[bi] = working_y[bi];
                  if (working_y[bi] > working_y_block_max[bi])
                    working_y_block_max[bi] = working_y[bi];
                }
  
                if (additional_output_file != NULL)
                {
                  fprintf(additional_output_file, "(Y) Multiple full decompression solutions within four blocks (%d, %d):\n", bx * 2, by * 2);
                  for (bi = 0; bi < 64; bi++)
                  {
                    if (bi % 8 == 0)
                      fprintf(additional_output_file, "\n");
                    fprintf(additional_output_file, "%d ", working_y[bi]);
                  }
                  fprintf(additional_output_file, "\n");
                }
              }
            }
    
            // Increment to the next possible block.
            for (bi_y = 0; bi_y < 64; bi_y++)
            {
              if (working_y[bi_y] < (INT16)refined_values.y.max[y_block_start + bi_y])
              {
                // Increment and test this one.
                working_y[bi_y]++;
                break;
              }
              else if (working_y[bi_y] == (INT16)refined_values.y.max[y_block_start + bi_y])
              {
                // Set this one back to zero and continue to carry.
                working_y[bi_y] = (INT16)refined_values.y.min[y_block_start + bi_y];
              }
            }
          }
  
          // Copy the hopefully refined sets back into the output data.
          for (bi = 0; bi < 64; bi++)
          {
  //          printf("Updating %d %d Cr %d %d to ", refined_values.cb.min[(by * refined_values.cb.width / 8 + bx) * 64 + bi], refined_values.cb.max[(by * refined_values.cb.width / 8 + bx) * 64 + bi], refined_values.cr.min[(by * refined_values.cr.width / 8 + bx) * 64 + bi], refined_values.cr.max[(by * refined_values.cr.width / 8 + bx) * 64 + bi]);
  
            refined_values.y.min[y_block_start + bi] = working_y_block_min[bi];
            refined_values.y.max[y_block_start + bi] = working_y_block_max[bi];
  
  //          printf("%d %d Cr %d %d\n", refined_values.cb.min[(by * refined_values.cb.width / 8 + bx) * 64 + bi], refined_values.cb.max[(by * refined_values.cb.width / 8 + bx) * 64 + bi], refined_values.cr.min[(by * refined_values.cr.width / 8 + bx) * 64 + bi], refined_values.cr.max[(by * refined_values.cr.width / 8 + bx) * 64 + bi]);
          }
  
          if (check_y1 && possibility_count(refined_values.y, bx * 2, by * 2) == 1L && y_element == 0)
          {
            y_block_status->ambiguous_block_count--;
            y_block_status->exact_block_count++;
          }
  
          if (check_y2 && possibility_count(refined_values.y, bx * 2 + 1, by * 2) == 1L && y_element == 1)
          {
            y_block_status->ambiguous_block_count--;
            y_block_status->exact_block_count++;
          }
  
          if (check_y3 && possibility_count(refined_values.y, bx * 2, by * 2 + 1) == 1L && y_element == 2)
          {
            y_block_status->ambiguous_block_count--;
            y_block_status->exact_block_count++;
          }
  
          if (check_y4 && possibility_count(refined_values.y, bx * 2 + 1, by * 2 + 1) == 1L && y_element == 3)
          {
            y_block_status->ambiguous_block_count--;
            y_block_status->exact_block_count++;
          }
        }
      }
    }
  }

  free(working_cb);
  free(working_cr);
  free(working_y1);
  free(working_y2);
  free(working_y3);
  free(working_y4);
  free(working_y);
  free(working_scratch_y);
  free(working_scratch_cb);
  free(working_scratch_cr);
  free(working_y_block_min);
  free(working_y_block_max);
  free(working_cb_block_min);
  free(working_cb_block_max);
  free(working_cr_block_min);
  free(working_cr_block_max);
  free(smoothed_block);
  destroy_interval_component_image_data(downsampled_spatial_data);
}

// Refine our estimates of luma and chroma quantized values, and returns whether refining was successful (i.e. there were no impossible blocks).
boolean refine_luma_and_chroma(block_interval_component_image_data values, block_interval_component_image_data refined_values, quantization_table luma, quantization_table chroma, interval_component_image_data* spatial_estimate, computation_plane smoothing_computations, inverse_ycc_rgb_table inverse_ycc_rgb_table, packed_image_data image, block_status_summary* y_block_status, block_status_summary* cb_block_status, block_status_summary* cr_block_status, boolean final_pass_output, const char* additional_output_file_name)
{
  FILE* additional_output_file = final_pass_output ? fopen(additional_output_file_name, "w") : NULL;
  if (additional_output_file != NULL)
    fprintf(additional_output_file, "Refine Y component\n");

  interval_plane new_y = refine_dct_plane(values.y, refined_values.y, luma, spatial_estimate->y, y_block_status, additional_output_file);
  if (additional_output_file != NULL)
    fprintf(additional_output_file, "Refine Cb component\n");

  interval_plane new_cb = refine_dct_plane(values.cb, refined_values.cb, chroma, spatial_estimate->cb, cb_block_status, additional_output_file);
  if (additional_output_file != NULL)
    fprintf(additional_output_file, "Refine Cr component\n");

  interval_plane new_cr = refine_dct_plane(values.cr, refined_values.cr, chroma, spatial_estimate->cr, cr_block_status, additional_output_file);
  destroy_interval_plane(spatial_estimate->y);
  destroy_interval_plane(spatial_estimate->cb);
  destroy_interval_plane(spatial_estimate->cr);
  spatial_estimate->y = new_y;
  spatial_estimate->cb = new_cb;
  spatial_estimate->cr = new_cr;

  if (y_block_status->impossible_block_count > 0 || cb_block_status->impossible_block_count > 0 || cr_block_status->impossible_block_count > 0)
  {
    if (additional_output_file != NULL)
      fclose(additional_output_file);
    return FALSE;
  }

  // Ambiguous blocks may be refined further by doing a full decompression check (rather than just checking whether spatial sample values are in the intervals).
  if (additional_output_file != NULL)
    fprintf(additional_output_file, "Full decompression refinement\n");
  full_decompression_refinement(refined_values, luma, chroma, smoothing_computations, inverse_ycc_rgb_table, image, y_block_status, cb_block_status, cr_block_status, additional_output_file);

  if (additional_output_file != NULL)
    fclose(additional_output_file);

  return TRUE;
}

void refine_upsampled_chroma(interval_component_image_data downsampled_chroma_estimate, set_plane upsampled_component_planes, computation_plane smoothing_computations) 
{
  UINT32 x, y, set_index, new_set_index;
  UINT8 cb, cr, yl;
  long int cb_min, cb_max, cr_min, cr_max, yl_min, yl_max;
  UINT32** new_sets;
  UINT32* set;
  UINT32* new_set;
  new_sets = (UINT32**)malloc(smoothing_computations.width * smoothing_computations.height * sizeof(UINT32*));

  // Upsample the chroma for Cb and Cr and use the results to narrow down the colour space conversion results.
  for (y = 0; y < smoothing_computations.height; y++)
  {
    for (x = 0; x < smoothing_computations.width; x++)
    {
      evaluate_computation(smoothing_computations.computations[y * smoothing_computations.width + x], &cb_max, &cb_min, downsampled_chroma_estimate.cb);
      evaluate_computation(smoothing_computations.computations[y * smoothing_computations.width + x], &cr_max, &cr_min, downsampled_chroma_estimate.cr);
      yl_min = downsampled_chroma_estimate.y.min[y * smoothing_computations.width + x];
      yl_max = downsampled_chroma_estimate.y.max[y * smoothing_computations.width + x];
      
      // Prune the set at this location so that only those elements which have a valid Cb and Cr remain.
      set = upsampled_component_planes.sets[y * smoothing_computations.width + x];
      new_set = new_sets[y * smoothing_computations.width + x] = (UINT32*)malloc(set[0] * sizeof(UINT32));
      new_set[0] = 1;
      new_set_index = 1;
      for (set_index = 1; set_index < set[0]; set_index++)
      {
        if (set[set_index] == LIST_SENTINEL32)
          break;
        yl = ((set[set_index] >> SHIFT_Y ) & 0x0000FF);
        cb = ((set[set_index] >> SHIFT_CB) & 0x0000FF);
        cr = ((set[set_index] >> SHIFT_CR) & 0x0000FF);

        if (!(cb < (UINT8)cb_min || cb > (UINT8)cb_max || cr < (UINT8)cr_min || cr > (UINT8)cr_max || yl < (UINT8)yl_min || yl > (UINT8)yl_max))
        {
          new_set[new_set_index++] = set[set_index];
          new_set[0]++;
        }
      }
      upsampled_component_planes.sets[y * smoothing_computations.width + x] = new_set;
    }
  }
}
