// Copyright (c) 2010 Andrew Lewis. All rights reserved.
// Use of this source code is governed the license in the LICENSE file.

#include "jpegout.h"

#undef INT8
#undef UINT8
#undef INT16
#undef UINT16
#undef INT32
#undef UINT32

#include <jpeglib.h>
#include <jerror.h>

#define INT8 char
#define UINT8 unsigned char
#define INT16 int
#define UINT16 unsigned int
#define INT32 int
#define UINT32 unsigned int
#define INT64 long int
#define UINT64 unsigned long int
#define INT128 __int128_t
#define UINT128 unsigned long long int

void jpeg_output(block_interval_component_image_data image, UINT8 quality_factor, const char* jpeg_file_name)
{
  struct jpeg_compress_struct cinfo;
  struct jpeg_error_mgr jerr;
  jvirt_barray_ptr* coeff_arrays = (jvirt_barray_ptr*)malloc(sizeof(jvirt_barray_ptr) * 3);
  UINT64 possibilities;
  UINT8 ci, bx, by, bi, interval_size;

  //printf("Y component values\n");

  for (by = 0; by < image.y.height / 8; by++)
  {
    for (bx = 0; bx < image.y.width / 8; bx++)
    {
      possibilities = 1L;
      for (bi = 0; bi < 64; bi++)
      {
        interval_size = image.y.max[(by * image.y.width / 8 + bx) * 64 + bi] - image.y.min[(by * image.y.width / 8 + bx) * 64 + bi];
        possibilities *= interval_size + 1;
      }
      //printf("Block %d %d has search size %d\n", bx, by, possibilities);
    }
  }

  //printf("End of Y\n");

  for (by = 0; by < image.cb.height / 8; by++)
  {
    for (bx = 0; bx < image.cb.width / 8; bx++)
    {
      possibilities = 1L;
      for (bi = 0; bi < 64; bi++)
      {
        interval_size = image.cb.max[(by * image.cb.width / 8 + bx) * 64 + bi] - image.cb.min[(by * image.cb.width / 8 + bx) * 64 + bi];
        possibilities *= interval_size + 1;
      }
      //printf("Block %d %d has search size %d\n", bx, by, possibilities);
    }
  }

  //printf("End of Cb\n");

  for (by = 0; by < image.cr.height / 8; by++)
  {
    for (bx = 0; bx < image.cr.width / 8; bx++)
    {
      possibilities = 1L;
      for (bi = 0; bi < 64; bi++)
      {
        interval_size = image.cr.max[(by * image.cr.width / 8 + bx) * 64 + bi] - image.cr.min[(by * image.cr.width / 8 + bx) * 64 + bi];
        possibilities *= interval_size + 1;
      }
      //printf("Block %d %d has search size %d\n", bx, by, possibilities);
    }
  }

  //printf("End of Cr\n");

  // Get ready to output to a file.
  FILE* outfile;
  if ((outfile = fopen(jpeg_file_name, "wb")) == NULL)
  {
    fprintf(stderr, "Can't open output JPEG file.\n");
    exit(1);
  }

  // Create compression object.
  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_compress(&cinfo);

  // Set all compression parameters as necessary.
  cinfo.image_width = image.y.width;
  cinfo.image_height = image.y.height;
  cinfo.input_components = 3;
  cinfo.in_color_space = JCS_RGB;
  jpeg_set_defaults(&cinfo);
  jpeg_set_quality(&cinfo, quality_factor, FALSE);
  jpeg_stdio_dest(&cinfo, outfile);
  cinfo.comp_info[0].h_samp_factor = 2;
  cinfo.comp_info[0].v_samp_factor = 2;
  cinfo.comp_info[1].h_samp_factor = 1;
  cinfo.comp_info[1].v_samp_factor = 1;
  cinfo.comp_info[1].h_samp_factor = 1;
  cinfo.comp_info[2].v_samp_factor = 1;

  // Make the block array descriptors.
  coeff_arrays[0] = (*cinfo.mem->request_virt_barray)((j_common_ptr)&cinfo, JPOOL_IMAGE, FALSE, image.y.width / 8, image.y.height / 8, (JDIMENSION)image.y.height * image.y.width);
  coeff_arrays[1] = (*cinfo.mem->request_virt_barray)((j_common_ptr)&cinfo, JPOOL_IMAGE, FALSE, image.cb.width / 8, image.cb.height / 8, (JDIMENSION)image.y.height * image.y.width);
  coeff_arrays[2] = (*cinfo.mem->request_virt_barray)((j_common_ptr)&cinfo, JPOOL_IMAGE, FALSE, image.cr.width / 8, image.cr.height / 8, (JDIMENSION)image.y.height * image.y.width);

  // Write out coefficients.
  jpeg_write_coefficients(&cinfo, coeff_arrays);

  // Fill virtual arrays.
  //printf("Fill virtual\n");
  for (ci = 0; ci < 3; ci++)
  {
    JBLOCKARRAY buffer;
    JCOEFPTR blockptr;
    jpeg_component_info* compptr;
    compptr = cinfo.comp_info + ci;
    for (by = 0; by < compptr->height_in_blocks; by++)
    {
      buffer = (cinfo.mem->access_virt_barray)((j_common_ptr)&cinfo, coeff_arrays[ci], by, (JDIMENSION)1, TRUE);
      for (bx = 0; bx < compptr->width_in_blocks; bx++)
      {
        blockptr = buffer[0][bx];
        for (bi = 0; bi < 64; bi++)
        {
          //printf("%d %d %d %d = %d\n", ci, by, bx, bi, ci == 0 ? image.y.max[(by * image.y.width / 8 + bx) * 64 + bi] : ci == 1 ? image.cb.max[(by * image.cb.width / 8 + bx) * 64 + bi] : image.cr.max[(by * image.cr.width / 8 + bx) * 64 + bi]);

          // For blocks where we haven't got an exact answer (ambiguous or infeasible), output a value between the minimum and maximum.
          if (ci == 0)
            blockptr[bi] = ((image.y.max[(by * image.y.width / 8 + bx) * 64 + bi] - image.y.min[(by * image.y.width / 8 + bx) * 64 + bi]) != 0 
              ? ((int)image.y.max[(by * image.y.width / 8 + bx) * 64 + bi] + (int)image.y.min[(by * image.y.width / 8 + bx) * 64 + bi]) >> 1
              : image.y.max[(by * image.y.width / 8 + bx) * 64 + bi]);
          else if (ci == 1)
            blockptr[bi] = ((image.cb.max[(by * image.cb.width / 8 + bx) * 64 + bi] - image.cb.min[(by * image.cb.width / 8 + bx) * 64 + bi]) != 0
              ? ((int)image.cb.max[(by * image.cb.width / 8 + bx) * 64 + bi] + (int)image.cb.min[(by * image.cb.width / 8 + bx) * 64 + bi]) >> 1
              : image.cb.max[(by * image.cb.width / 8 + bx) * 64 + bi]);
          else if (ci == 2)
            blockptr[bi] = ((image.cr.max[(by * image.cr.width / 8 + bx) * 64 + bi] - image.cr.min[(by * image.cr.width / 8 + bx) * 64 + bi]) != 0
              ? ((int)image.cr.max[(by * image.cr.width / 8 + bx) * 64 + bi] + (int)image.cr.min[(by * image.cr.width / 8 + bx) * 64 + bi]) >> 1
              : image.cr.max[(by * image.cr.width / 8 + bx) * 64 + bi]);
        }
      }
    }
  }
  //printf("Done\n");

  // jpeg_finish_compress
  jpeg_finish_compress(&cinfo);

  // Destroy compression object.
  jpeg_destroy_compress(&cinfo);

  fclose(outfile);
  free(coeff_arrays);
}
