// Copyright (c) 2010 Andrew Lewis. All rights reserved.
// Use of this source code is governed the license in the LICENSE file.

#include "unsmooth.h"

computation_plane get_smoothing_computations_plane(set_plane component_sets)
{
  // TODO: Here we assume a multiple of two for the upsampled image width and height.
  computation_plane computations_plane = create_computation_plane(component_sets, component_sets.width / 2, component_sets.height / 2);
  Computation** current_computation_index = computations_plane.computations;
  Computation* cThiscolsum; Computation* cLastcolsum; Computation* cNextcolsum;
  UINT16 v, x, y;

  for (y = 0; y < computations_plane.input_height; y++)
  {
    // Two rows are created for each input row.
    for (v = 0; v < 2; v++)
    {
      cThiscolsum = create_computation(COMPUTATION_ADD,
                                    create_computation(COMPUTATION_INTEGER_MULTIPLY,
                                                    create_computation(COMPUTATION_INPUT_JSAMPLE, NULL, NULL, 0, y * computations_plane.input_width), NULL,
                                                    3, 0),
                                    create_computation(COMPUTATION_INPUT_JSAMPLE, NULL, NULL, 0, y * computations_plane.input_width + (v == 0 ? -computations_plane.input_width : computations_plane.input_width)), 0, 0);
  
      cNextcolsum = create_computation(COMPUTATION_ADD,
                                    create_computation(COMPUTATION_INTEGER_MULTIPLY,
                                                    create_computation(COMPUTATION_INPUT_JSAMPLE, NULL, NULL, 0, y * computations_plane.input_width + 1), NULL,
                                                    3, 0),
                                    create_computation(COMPUTATION_INPUT_JSAMPLE, NULL, NULL, 0, y * computations_plane.input_width + (v == 0 ? -computations_plane.input_width : computations_plane.input_width) + 1), 0, 0);
  
      *(current_computation_index++) = create_computation(COMPUTATION_INTEGER_RSHIFT,
                          create_computation(COMPUTATION_INTEGER_ADD,
                                          create_computation(COMPUTATION_INTEGER_MULTIPLY, clone_computation(cThiscolsum), NULL, 4, 0),
                                          NULL, 8, 0),
                          NULL, 4, 0); // References cThiscolsum.
      *(current_computation_index++) = create_computation(COMPUTATION_INTEGER_RSHIFT,
                          create_computation(COMPUTATION_INTEGER_ADD,
                                          create_computation(COMPUTATION_ADD,
                                                          create_computation(COMPUTATION_INTEGER_MULTIPLY, clone_computation(cThiscolsum), NULL, 3, 0),
                                                          clone_computation(cNextcolsum), 0, 0),
                                          NULL, 7, 0),
                          NULL, 4, 0); // References cNextcolsum and a clone of cThiscolsum.
  
      cLastcolsum = cThiscolsum; cThiscolsum = cNextcolsum;
  
      for (x = 2; x < computations_plane.input_width; x++)
      {
        // General case: 3/4 * nearer pixel + 1/4 * further pixel in each dimension, thus 9/16, 3/16, 3/16, 1/16 overall.
        cNextcolsum = create_computation(COMPUTATION_ADD,
                                      create_computation(COMPUTATION_INTEGER_MULTIPLY,
                                                      create_computation(COMPUTATION_INPUT_JSAMPLE, NULL, NULL, 0, y * computations_plane.input_width + x), NULL,
                                                      3, 0),
                                      create_computation(COMPUTATION_INPUT_JSAMPLE, NULL, NULL, 0, y * computations_plane.input_width + x + (v == 0 ? -computations_plane.input_width : computations_plane.input_width)), 0, 0);
        *(current_computation_index++) = create_computation(COMPUTATION_INTEGER_RSHIFT,
                            create_computation(COMPUTATION_INTEGER_ADD,
                                            create_computation(COMPUTATION_ADD,
                                                            create_computation(COMPUTATION_INTEGER_MULTIPLY, clone_computation(cThiscolsum), NULL, 3, 0),
                                                            cLastcolsum, 0, 0),
                                            NULL, 8, 0),
                            NULL, 4, 0); // References cThiscolsum and cLastcolsum.
        *(current_computation_index++) = create_computation(COMPUTATION_INTEGER_RSHIFT,
                            create_computation(COMPUTATION_INTEGER_ADD,
                                            create_computation(COMPUTATION_ADD,
                                                            create_computation(COMPUTATION_INTEGER_MULTIPLY, clone_computation(cThiscolsum), NULL, 3, 0),
                                                            clone_computation(cNextcolsum), 0, 0),
                                            NULL, 7, 0),
                            NULL, 4, 0); // References cNextcolsum and a clone of cThiscolsum.

        cLastcolsum = cThiscolsum; cThiscolsum = cNextcolsum;
      }
      // Last two computations
      *(current_computation_index++) = create_computation(COMPUTATION_INTEGER_RSHIFT,
                          create_computation(COMPUTATION_INTEGER_ADD,
                                          create_computation(COMPUTATION_ADD,
                                                          create_computation(COMPUTATION_INTEGER_MULTIPLY, clone_computation(cThiscolsum), NULL, 3, 0),
                                                          cLastcolsum, 0, 0),
                                          NULL, 8, 0),
                          NULL, 4, 0); // References cThiscolsum.
      *(current_computation_index++) = create_computation(COMPUTATION_INTEGER_RSHIFT,
                          create_computation(COMPUTATION_INTEGER_ADD,
                                          create_computation(COMPUTATION_INTEGER_MULTIPLY, cThiscolsum, NULL, 4, 0),
                                          NULL, 7, 0),
                          NULL, 4, 0); // References cNextcolsum and a clone of cThiscolsum.
    }
  }
  
  return computations_plane;
}
